from pathlib import Path
import yaml

from biolib import cli_utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


def push_application(author_and_app_name: str, app_path: str):
    api_client = BiolibApiClient.get()
    if not api_client.is_signed_in:
        # TODO: Create an exception class for expected errors like this that does not print stacktrace
        raise Exception(
            'You must be authenticated to push an application.\n'
            'Please set the environment variable "BIOLIB_TOKEN=[your_api_token]"\n'
            'You can get an API token at: https://biolib.com/settings/api-tokens/'
        ) from None

    # prepare zip file
    if not Path(f'{app_path}/.biolib/config.yml').is_file():
        raise BioLibError('The file .biolib/config.yml was not found in the application directory')

    tmp_directory, source_files_zip_path = cli_utils.create_temporary_directory_with_source_files_zip(app_path)

    try:
        with open(source_files_zip_path, mode='rb') as zip_file:
            source_files_zip_bytes = zip_file.read()
    except IOError:
        raise Exception('Failed to read temporary source files zip') from None

    try:
        tmp_directory.cleanup()
    except IOError:
        raise Exception('Failed to clean up temporary source files zip directory') from None

    author, app_name = author_and_app_name.split('/')
    app = BiolibAppApi.fetch_by_name(author, app_name)
    # push new app version
    new_app_version_json = BiolibAppApi.push_app_version(
        app_id=app['public_id'],
        zip_binary=source_files_zip_bytes,
        author=author,
        app_name=app_name,
        set_as_active=False
    )

    docker_tags = new_app_version_json.get('docker_tags', {})
    if docker_tags:
        logger.info('Found docker images to push.')

        try:
            yaml_file = open(f'{app_path}/.biolib/config.yml', 'r', encoding='utf-8')

        except Exception as error:  # pylint: disable=broad-except
            raise BioLibError('Could not open the config file .biolib/config.yml') from error

        try:
            config_data = yaml.safe_load(yaml_file)

        except Exception as error:  # pylint: disable=broad-except
            raise BioLibError('Could not parse .biolib/config.yml. Please make sure it is valid YAML') from error

        # Auth to be sent to proxy
        # The tokens are sent as "{access_token},{job_id}". We leave job_id blank on push.
        tokens = f'{BiolibApiClient.get().access_token},'
        auth_config = {'username': 'biolib', 'password': tokens}

        docker_client = BiolibDockerClient.get_docker_client()

        for module_name, repo_and_tag in docker_tags.items():
            docker_image_definition = config_data['modules'][module_name]['image']
            repo, tag = repo_and_tag.split(':')

            if docker_image_definition.startswith('dockerhub://'):
                docker_image_name = docker_image_definition.replace('dockerhub://', 'docker.io/', 1)
                logger.info(
                    f'Pulling image {docker_image_name} defined on module {module_name} from Dockerhub.')
                dockerhub_repo, dockerhub_tag = docker_image_name.split(':')
                docker_client.images.pull(dockerhub_repo, tag=dockerhub_tag)

            elif docker_image_definition.startswith('local-docker://'):
                docker_image_name = docker_image_definition.replace('local-docker://', '', 1)

            try:
                logger.info(f'Trying to push image {docker_image_name} defined on module {module_name}.')
                image = docker_client.images.get(docker_image_name)
                ecr_proxy_host = f'containers.{api_client.base_url.split("://")[1]}/{repo}'
                image.tag(ecr_proxy_host, tag)
                for line in docker_client.images.push(ecr_proxy_host, tag=tag, stream=True,
                                                      decode=True, auth_config=auth_config):
                    logger.info(line)

            except Exception as exception:  # pylint: disable=broad-except
                raise BioLibError(f'Failed to tag and push image {docker_image_name}.') from exception

            logger.info(f'Successfully pushed {docker_image_name}')

        logger.info('Successfully pushed all docker images')

    # Set new app version as active
    BiolibAppApi.update_app_version(
        app_version_id=new_app_version_json['public_id'],
        data={
            'set_as_active': True
        }
    )



import abc

from biolib.compute_node.job_worker.executors.types import LocalExecutorOptions


class BaseExecutor(abc.ABC):

    def __init__(self, options: LocalExecutorOptions):
        self._options: LocalExecutorOptions = options

    @abc.abstractmethod
    def execute_module(self, module_input_serialized: bytes) -> bytes:
        pass



import time
from datetime import datetime
import requests

from biolib import utils
from biolib.typing_utils import Dict, List, Optional
from biolib.biolib_api_client import BiolibApiClient
from biolib.compute_node.webserver.worker_thread import WorkerThread
from biolib.compute_node.webserver import webserver_config
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.compute_node import enclave
from biolib.compute_node.enclave.parent_api_types import ComputeNodeInfo

JOB_ID_TO_COMPUTE_STATE_DICT: Dict = {}
UNASSIGNED_COMPUTE_PROCESSES: List = []


def finalize_and_clean_up_compute_job(job_id: str):
    JOB_ID_TO_COMPUTE_STATE_DICT.pop(job_id)

    if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE and not utils.IS_DEV:
        logger.debug('Deregistering and shutting down...')
        enclave.ParentApi.deregister_and_shutdown()


def get_compute_state(unassigned_compute_processes):
    if len(unassigned_compute_processes) == 0:
        start_compute_process(unassigned_compute_processes)

    return unassigned_compute_processes.pop()


def start_compute_process(unassigned_compute_processes):
    compute_state = {
        'job_id': None,
        'status': {
            'status_updates': [
                {
                    'progress': 10,
                    'log_message': 'Initializing'
                }
            ],
        },
        'result': None,
        'attestation_document': b'',
        'received_messages_queue': None,
        'messages_to_send_queue': None,
        'worker_process': None
    }

    WorkerThread(compute_state).start()

    while True:
        if compute_state['attestation_document']:
            break
        time.sleep(1)

    unassigned_compute_processes.append(compute_state)


def validate_saved_job(saved_job):
    if 'app_version' not in saved_job['job']:
        return False

    if 'access_token' not in saved_job:
        return False

    if 'module_name' not in saved_job:
        return False

    return True


def report_availability(compute_node_info: ComputeNodeInfo):
    try:
        api_client = BiolibApiClient.get()
        logger.debug(f'Registering with {compute_node_info} to host {api_client.base_url} at {datetime.now()}')

        response: Optional[requests.Response] = None
        max_retries = 5
        for retry_count in range(max_retries):
            try:
                response = requests.post(f'{api_client.base_url}/api/jobs/report_available/', json=compute_node_info)
                break
            except Exception as error:  # pylint: disable=broad-except
                logger.error(f'Self-registering failed with error: {error}')
                if retry_count < max_retries - 1:
                    seconds_to_sleep = 1
                    logger.info(f'Retrying self-registering in {seconds_to_sleep} seconds')
                    time.sleep(seconds_to_sleep)

        if not response:
            raise BioLibError('Failed to register. Max retry limit reached')

        if response.status_code != 201:
            raise Exception("Non 201 error code")

        if response.json()['is_reserved']:
            # Start running job shutdown timer if reserved. It restarts when the job is actually saved
            enclave.ParentApi.start_shutdown_timer(webserver_config.COMPUTE_NODE_RUNNING_JOB_SHUTDOWN_TIME_MINUTES)

        else:
            # Else start the longer auto shutdown timer
            enclave.ParentApi.start_shutdown_timer(webserver_config.COMPUTE_NODE_AUTO_SHUTDOWN_TIME_MINUTES)

    except Exception as exception:  # pylint: disable=broad-except
        logger.error(f'Shutting down as self register failed due to: {exception}')
        if not utils.IS_DEV:
            enclave.ParentApi.deregister_and_shutdown()



import json
import time
import base64
import logging
from flask import Flask, request, Response, jsonify
from flask_cors import CORS  # type: ignore

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_binary_format import SavedJob
from biolib.compute_node.enclave.enclave_remote_hosts import start_enclave_remote_hosts
from biolib.compute_node.webserver import webserver_utils
from biolib.compute_node.webserver import webserver_config
from biolib.compute_node.webserver.gunicorn_flask_application import GunicornFlaskApplication
from biolib.biolib_logging import logger, TRACE
from biolib.compute_node import enclave

app = Flask(__name__)
CORS(app)


@app.route('/hello/')
def hello():
    return 'Hello'


@app.route('/v1/job/', methods=['POST'])
def save_job():
    saved_job = json.loads(request.data.decode())

    # TODO: figure out why this shallow validate method is used
    if not webserver_utils.validate_saved_job(saved_job):
        return jsonify({'job': 'Invalid job'}), 400

    job_id = saved_job['job']['public_id']
    saved_job['BASE_URL'] = BiolibApiClient.get().base_url

    compute_state = webserver_utils.get_compute_state(webserver_utils.UNASSIGNED_COMPUTE_PROCESSES)
    compute_state['job_id'] = job_id
    webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id] = compute_state

    if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
        enclave_config = enclave.ParentApi.get_enclave_config()
        saved_job['biolib_containers_pull_token'] = enclave_config['biolib_containers_pull_token']
        if not utils.IS_DEV:
            enclave.ParentApi.start_shutdown_timer(webserver_config.COMPUTE_NODE_RUNNING_JOB_SHUTDOWN_TIME_MINUTES)

    saved_job_bbf_package = SavedJob().serialize(json.dumps(saved_job))
    send_package_to_compute_process(job_id, saved_job_bbf_package)

    if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
        return Response(base64.b64encode(compute_state['attestation_document']), status=201)
    else:
        return '', 201


@app.route('/v1/job/<job_id>/start/', methods=['POST'])
def start_compute(job_id):
    module_input_package = request.data
    send_package_to_compute_process(job_id, module_input_package)
    return '', 201


@app.route('/v1/job/<job_id>/status/')
def status(job_id):
    # TODO Implement auth token
    current_status = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status'].copy()
    response = jsonify(current_status)

    if current_status['status_updates']:
        webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status']['status_updates'] = []

    # Check if any error occurred
    if 'error_code' in current_status:
        response.call_on_close(lambda: webserver_utils.finalize_and_clean_up_compute_job(job_id))

    return response


@app.route('/v1/job/<job_id>/result/')
def result(job_id):
    if webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['result']:
        result_data = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['result']
        # remove result from state dict, so we know the user has started the download
        webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['result'] = None
        response = Response(result_data)
        response.call_on_close(lambda: webserver_utils.finalize_and_clean_up_compute_job(job_id))

        return response

    else:
        return '', 404


def send_package_to_compute_process(job_id, package_bytes):
    message_queue = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['messages_to_send_queue']
    message_queue.put(package_bytes)


def start_webserver(port, host):
    def worker_exit(server, worker):  # pylint: disable=unused-argument
        active_compute_states = list(
            webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT.values()) + webserver_utils.UNASSIGNED_COMPUTE_PROCESSES
        logger.debug(f'Sending terminate signal to {len(active_compute_states)} compute processes')
        if active_compute_states:
            for compute_state in active_compute_states:
                if compute_state['worker_thread']:
                    compute_state['worker_thread'].terminate()
            time.sleep(2)
        return

    def post_fork(server, worker):  # pylint: disable=unused-argument
        logger.info('Started compute node')
        webserver_utils.start_compute_process(webserver_utils.UNASSIGNED_COMPUTE_PROCESSES)

        if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            logger.debug('Initializing enclave')
            config = enclave.ParentApi.get_enclave_config()
            utils.IS_DEV = config['is_dev']
            BiolibApiClient.initialize(config['base_url'])
            start_enclave_remote_hosts()
            logging.debug('Enclave initialized. Reporting availability...')
            webserver_utils.report_availability(config['compute_node_info'])

    if logger.level == TRACE:
        gunicorn_log_level_name = 'DEBUG'
    elif logger.level == logging.DEBUG:
        gunicorn_log_level_name = 'INFO'
    elif logger.level == logging.INFO:
        gunicorn_log_level_name = 'WARNING'
    else:
        gunicorn_log_level_name = logging.getLevelName(logger.level)

    options = {
        'bind': f'{host}:{port}',
        'workers': 1,
        'post_fork': post_fork,
        'worker_exit': worker_exit,
        'timeout': webserver_config.GUNICORN_REQUEST_TIMEOUT,
        'graceful_timeout': 4,
        'loglevel': gunicorn_log_level_name,
    }

    GunicornFlaskApplication(app, options).run()
